/*
 * $Id$
 * 
 *
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *  
 *
 * @author <a href="mailto:pbwest@powerup.com.au">Peter B. West</a>
 */

package org.apache.fop.fo.flow;

// FOP
import java.util.Arrays;
import java.util.BitSet;

import org.apache.fop.apps.FOPException;
import org.apache.fop.datastructs.TreeException;
import org.apache.fop.fo.FONode;
import org.apache.fop.fo.FOPageSeqNode;
import org.apache.fop.fo.FOTree;
import org.apache.fop.fo.FObjectNames;
import org.apache.fop.fo.PropNames;
import org.apache.fop.fo.PropertySets;
import org.apache.fop.fo.expr.PropertyException;
import org.apache.fop.xml.FoXmlEvent;
import org.apache.fop.xml.XmlEvent;
import org.apache.fop.xml.XmlEventReader;

/**
 * Implements the fo:footnote flow object.
 * <p>Footnote is an extremely messy flow object.  The only absolute
 * prohibition seem sto be that it may not be a descendant of another
 * fo:footnote.  <b>6.10.3 fo:footnote</b> <i>Constraints</i> states:
* <p><tt>"It is an error if the fo:footnote occurs as a
 * descendant of a flow that is not assigned to a region-body, or of an
 * fo:block-container that generates absolutely positioned areas. In either
 * case, the block-areas generated by the fo:footnote-body child of the
 * fo:footnote shall be returned to the parent of the fo:footnote and placed
 * in the area tree as though they were normal block-level areas."</tt>
 * <p>In other words, it's wrong, but you can do it anyway.  The end result
 * is that fo:footnote can appear almost anywhere.
 */
public class FoFootnote extends FOPageSeqNode {

    private static final String tag = "$Name$";
    private static final String revision = "$Revision$";

    /** Map of <tt>Integer</tt> indices of <i>sparsePropsSet</i> array.
        It is indexed by the FO index of the FO associated with a given
        position in the <i>sparsePropsSet</i> array.
     */
    private static final int[] sparsePropsMap;

    /** An <tt>int</tt> array of of the applicable property indices, in
        property index order. */
    private static final int[] sparseIndices;

    /** The number of applicable properties.  This is the size of the
        <i>sparsePropsSet</i> array. */
    private static final int numProps;

    static {
        // Collect the sets of properties that apply
        BitSet propsets = new BitSet();
        propsets.or(PropertySets.accessibilitySet);

        // Map these properties into sparsePropsSet
        // sparsePropsSet is a HashMap containing the indicies of the
        // sparsePropsSet array, indexed by the FO index of the FO slot
        // in sparsePropsSet.
        sparsePropsMap = new int[PropNames.LAST_PROPERTY_INDEX + 1];
        Arrays.fill(sparsePropsMap, -1);
        numProps = propsets.cardinality();
        sparseIndices = new int[numProps];
        int propx = 0;
        for (int next = propsets.nextSetBit(0);
                next >= 0;
                next = propsets.nextSetBit(next + 1)) {
            sparseIndices[propx] = next;
            sparsePropsMap[next] = propx++;
        }
    }

    /**
     * Construct an fo:footnote node, and build the fo:footnote subtree.
     * <p>Content model for fo:footnote: (inline,footnote-body)
     * @param foTree the FO tree being built
     * @param pageSequence ancestor of this node
     * @param parent the parent FONode of this node
     * @param event the <tt>XmlEvent</tt> that triggered the creation of
     * this node
     * @param stateFlags - passed down from the parent.  Includes the
     * attribute set information.
     */
    public FoFootnote
            (FOTree foTree, FoPageSequence pageSequence, FOPageSeqNode parent,
                    FoXmlEvent event, int stateFlags)
        throws TreeException, FOPException
    {
        super(foTree, FObjectNames.FOOTNOTE, pageSequence, parent, event,
                          stateFlags, sparsePropsMap, sparseIndices);
        if ((stateFlags & FONode.MC_FOOTNOTE) != 0)
            throw new FOPException
                    ("fo:footnote not allowed as child of fo:footnote.");
        XmlEvent ev;
        try {
            // Look for the inline
            if ((ev = xmlevents.expectStartElement
                    (FObjectNames.INLINE, XmlEvent.DISCARD_W_SPACE))
                   == null)
                throw new FOPException("No inline in footnote.");
            new FoInline(
                    getFOTree(), pageSequence, this, (FoXmlEvent)ev,
                    stateFlags | FONode.MC_FOOTNOTE);
            ev =  xmlevents.getEndElement(
                    XmlEventReader.DISCARD_EV, ev);
            namespaces.relinquishEvent(ev);

            // Look for the footnote-body
            if ((ev = xmlevents.expectStartElement
                    (FObjectNames.FOOTNOTE_BODY, XmlEvent.DISCARD_W_SPACE))
                   == null)
                throw new FOPException("No footnote-body in footnote.");
            new FoFootnoteBody(
                    getFOTree(), pageSequence, this, (FoXmlEvent)ev,
                    stateFlags | FONode.MC_FOOTNOTE);
            ev = xmlevents.getEndElement(XmlEventReader.DISCARD_EV, ev);
            namespaces.relinquishEvent(ev);

            /*
        } catch (NoSuchElementException e) {
            throw new FOPException
                ("Unexpected EOF while processing " + nowProcessing + ".");
            */
        } catch(TreeException e) {
            throw new FOPException("TreeException: " + e.getMessage());
        } catch(PropertyException e) {
            throw new FOPException("PropertyException: " + e.getMessage());
        }

        makeSparsePropsSet();
    }

}
